#!C:\Python24\python.exe
# a simple implementation of a dbgp proxy
#
# Copyright (c) 2003 ActiveState Corporation.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# Authors:
#    Shane Caraveo <ShaneC@ActiveState.com>
#    Trent Mick <TrentM@ActiveState.com>


import getopt, string, os, time, socket, select, threading, sys
import dbgp.serverBase
from dbgp.common import *

import cPickle # to get around a bug in the Installer wrapping
import logging
log = logging.getLogger("dbgp.proxy")
IDENTCHARS = string.letters + string.digits + '_'

# the host class implements commands that the host can send
# to the debugger target
class sessionProxy(dbgp.serverBase.session):
    def __init__(self, sessionHost):
        dbgp.serverBase.session.__init__(self, sessionHost)
        self._server = None
        self._serverAddr = None
        self._srvthread = None

    def __del__(self):
        if self._socket:
            self._socket.close()
        if self._server:
            self._server.close()
        log.debug("destructor [%r]", self)

    def _handleInitPacket(self):
        log.debug("session getting data")
        data = self._socket.recv(1024)
        eop = data.find('\0')
        try:
            size = long(data[:eop])
        except:
            # invalid protocol, close the socket
            log.info("%r Closing connection, invalid protocol used", self)
            self._socket.close()
            self.stop()
            return
        data = data[eop+1:] # skip \0
        sizeLeft = size - len(data) + 1
        while sizeLeft > 0:
            log.debug("session getting more data size=%d", sizeLeft)
            newdata = self._socket.recv(sizeLeft)
            data = data + newdata
            sizeLeft = sizeLeft - len(newdata)
        
        response = data[:size]
        data = data[size+1:] # skip \0
        log.debug("session dispatching call")
        self._dispatch(size,response)

    def start(self, socket, clientAddr):
        if not self._sessionHost.onConnect(self, socket, clientAddr):
            socket.close()
            return 0
        self._socket = socket
        self._clientAddr = clientAddr
        # create a new thread and initiate a debugger session
        if not self._cmdthread or not self._cmdthread.isAlive():
            self._cmdthread = threading.Thread(target = self._cmdloop)
            self._cmdthread.start()
        return 1

    def _serverLoop(self):
        try:
            while self._server and not self._stop:
                try:
                    data = self._server.recv(8192)
                    if not data:
                        log.debug("received no data from server, stopping")
                        self._stop = 1
                        break
                    self._socket.send(data)
    
                    if data.startswith('stop '):
                        self._stop = 1
                except Exception, e:
                    log.exception("session Exception in _proxySock [%r]", self)
                    self._server = None
                    self._stop = 1
                    break
        finally:
            self._srvthread = None
        
    def _cmdloop(self):
        # first thing we expect is an init packet from the client
        self._handleInitPacket()
        if self._stop:
            return

        # start our other proxy thread
        if not self._srvthread or not self._srvthread.isAlive():
            self._srvthread = threading.Thread(target = self._serverLoop)
            self._srvthread.start()
        
        # now we simply pass packets back and forth
        try:
            while self._socket and not self._stop:
                try:
                    data = self._socket.recv(8192)
                    if not data:
                        log.debug("received no data from client, stopping")
                        self._socket.close()
                        self._socket = None
                        self._stop = 1
                        break
                    if not self._server:
                        log.debug("got data from client, no server: %s", data)
                        self._socket.close()
                        self._socket = None
                        self._stop = 1
                        break
                    self._server.send(data)
                except Exception, e:
                    log.exception("session Exception in _cmdloop [%r]", self)
                    self._socket = None
                    self._stop = 1
                    break
        finally:
            self._cmdthread = None
    
    def stop(self):
        dbgp.serverBase.session.stop(self)
        self._stop = 1
        self._socket.close()
        self._socket = None
        log.info('session stopped')
    
    def startServer(self, server, initNode):
        self._serverAddr = server[0]
        try:
            self._server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._server.connect((self._serverAddr[0], self._serverAddr[1]))
        except socket.error, details:
            log.exception("Unable to connect to the server listener %s:%d [%r]",
                self._serverAddr[0],self._serverAddr[1], self)
            self._server = None
            self._serverAddr = None
            return 0
        # pass the init packet through
        response = initNode.toxml()
        l = len(response)
        log.debug("Sending response %d\0%s\0", l, response)
        self._server.send('%d\0%s\0' % (l, response))
        return 1

class clientHandler(dbgp.serverBase.listener):
    def __init__(self, manager, addr='127.0.0.1', port=9000):
        dbgp.serverBase.listener.__init__(self, manager)
        self.start(addr, port)

    def startNewSession(self, client, addr):
        # start a new thread that is the host connection
        # for this debugger session
        session = sessionProxy(self._session_host)
        session.start(client, addr)

class serverHandler(dbgp.serverBase.listener):
    def __init__(self, manager, addr='127.0.0.1', port=9001, caddr='127.0.0.1', cport=9000):
        self.manager = manager
        self.caddr = caddr
        self.cport = cport
        dbgp.serverBase.listener.__init__(self, self)
        self.start(addr, port)

    def parseline(self, line):
        line = line.strip()
        if not line:
            return None, None, line
        elif line[0] == '?':
            line = 'help ' + line[1:]
        elif line[0] == '!':
            if hasattr(self, 'do_shell'):
                line = 'shell ' + line[1:]
            else:
                return None, None, line
        i, n = 0, len(line)
        while i < n and line[i] in IDENTCHARS: i = i+1
        cmd, arg = line[:i], line[i:].strip()
        arg = arg.split()
        return cmd, arg, line

    def startNewSession(self, client, addr):
        # before any communication, we can decide if we want
        # to allow the connection here.  return 0 to deny
        data = client.recv(1024)
        log.info("Server:onConnect %r [%s]", addr, data)
        command, args, line = self.parseline(data)
        func = 'do_' + command
        try:
            func = getattr(self, 'do_' + command)
        except AttributeError, e:
            self._error(client, addr, command, str(e))
        try:
            func(client, addr, args)
        except Exception, e:
            self._error(client, addr, command, str(e))
        client.close()
        return 0
    
    def do_proxyinit(self, client, addr, args):
        opts, args = getopt.getopt(args, 'p:k:m:')
        port = idekey = multi = None
        for o, a in opts:
            if o == '-p':
                port = long(a)
            elif o == '-k':
                idekey = a
            elif o == '-m':
                multi = a
        if not idekey or not port:
            self._error(client,addr,'proxyinit','No port defined for proxy')
            return
        addr = [addr[0],port]
        id = self.manager.addServer(idekey, addr, multi)
        if id:
            client.send('<proxyinit success="1" idekey="%s" address="%s" port="%d"/>' %(id, self.caddr, self.cport))
        else:
            self._error(client,addr,'proxyinit','IDE Key already exists')

    def do_proxystop(self, client, addr, args):
        opts, args = getopt.getopt(args, 'k:')
        idekey = None
        for o, a in opts:
            if o == '-k':
                idekey = a
        if not idekey:
            self._error(client,addr,'proxystop','No IDE key')
            return
        id = self.manager.removeServer(idekey)
        client.send('<proxystop success="1" idekey="%s"/>' % id)

    def _error(self, client, addr, command, message):
        _err = '<%s success="0"><error id="%d"><message>%s</message></error></%s>'     
        client.send(_err % (command,0,message,command))

class proxy:
    _servers = {}
    
    def __init__(self, caddr='127.0.0.1', cport=9000,
                       saddr='127.0.0.1', sport=9001):
        log.info("starting proxy listeners.  appid: %d", os.getpid())
        log.info("    dbgp listener on %s:%d", caddr, cport)
        log.info("    IDE listener on  %s:%d", saddr, sport)
        self._clientListener = clientHandler(self, caddr, cport)
        self._serverListener = serverHandler(self, saddr, sport, caddr, cport)
    
    def addServer(self, idekey, addr, multi):
        if self._servers.has_key(idekey):
            return None
        self._servers[idekey] = [addr, multi]
        return idekey

    def removeServer(self, idekey):
        if self._servers.has_key(idekey):
            del self._servers[idekey]
            return idekey
        return None

    def getServer(self, idekey):
        if self._servers.has_key(idekey):
            return self._servers[idekey]
        return None
    
    def waitForever(self):
        try:
            while 1:
                time.sleep(1)
        except KeyboardInterrupt:
            self._clientListener.stop()
            self._serverListener.stop()

    ##################################################################
    # sessionProxy callback functions
    ##################################################################

    def onConnect(self, session, client, addr):
        # here we can handle security, return zero to deny
        # the connection.
        log.info("connection from %s:%d [%r]", addr[0], addr[1], session)
        return 1
    
    def initHandler(self, session, init):
        # this is called once during a session, after the connection
        # to provide initialization information.  initNode is a
        # minidom node.
        # 1. find the server the client wants
        # 2. add data to the init node
        # 3. call session.startServer
        idekey = init.getAttribute('idekey')
        server = self.getServer(idekey)
        if not server:
            log.warn("No server with key [%s], stopping request [%r]", idekey, session)
            session.stop()
            return
        init.setAttribute('proxied','true')
        if not session.startServer(server, init):
            log.warn("Unable to connect to server with key [%s], stopping request [%r]", idekey, session)
            self.removeServer(idekey)
            session.stop()
        
    
    def outputHandler(self, session, stream, text):
        # We should never get called
        raise "%r Ouch, the proxy messed up!" % session

    def responseHandler(self, session, response):
        # We should never get called
        raise "%r Ouch, the proxy messed up!" % session

__doc__ = """
    pydbgpproxy -i localhost:9001 -d localhost:9000
    
    -d hostname:port  listener port for debug processes
    -i hostname:port  listener port for IDE processes
    -h                help
    -l log_level      Logging levels from the logging module:
                        CRITICAL
                        ERROR
                        WARN
                        INFO
                        DEBUG

    The proxy listens on two ports, one for debugger session
    requests, and one for notifications from IDE's or other
    debugger front end tools.  This allows multiuser systems
    to provide a well defined port for debugger engines, while
    each front end would be listening in a unique port.

"""

__version__ = "$Revision: #1 $ $Change: 118914 $"
_version_ = [1, 0, 0]

def main(args):
    import sys
    if sys.version_info < (2, 0):
        exe = sys.executable
        ver = '.'.join(map(str, sys.version_info[:3]))
        sys.stderr.write("The proxy must be run with Python "
                         "version 2.0 or greater. Your current "
                         "python, '%s', is version '%s'\n"
                         % (exe, ver))
        sys.exit(1)
    
    try:
        optlist, args = getopt.getopt(args, 'hVd:l:i:',
            ['help', 'version', 'debug_port',
             'log_level', 'ide_port'])
    except getopt.GetoptError, msg:
        sys.stderr.write("proxy: error: %s\n" % str(msg))
        sys.stderr.write("See 'proxy --help'.\n")
        return 1

    idehost = dbghost = '127.0.0.1'
    dbgport = 9000
    ideport = 9001
    logLevel = logging.INFO

    for opt, optarg in optlist:
        if opt in ('-h', '--help'):
            sys.stdout.write(__doc__)
            return 0
        elif opt in ('-V', '--version'):
            import re
            kw = re.findall('\$(\w+):\s(.*?)\s\$', __version__)
            
            sys.stderr.write("pydbgpproxy Version %s %s %s %s %s\n"\
                             % ('.'.join([str(i) for i in _version_]),
                                kw[0][0], kw[0][1], kw[1][0], kw[1][1]))
            return 0
        elif opt in ('-d', '--debug_port'):
            if optarg.find(':') >= 0:
                dbghost, dbgport = optarg.split(':')
                dbgport = int(dbgport)
            else:
                dbghost = '127.0.0.1'
                dbgport = int(optarg)
        elif opt in ('-i', '--ide_port'):
            if optarg.find(':') >= 0:
                idehost, ideport = optarg.split(':')
                ideport = int(ideport)
            else:
                idehost = '127.0.0.1'
                ideport = int(optarg)
        elif opt in ('-l', '--log_level'):
            if optarg in logging._levelNames:
                logLevel = logging._levelNames[optarg]
            else:
                sys.stderr.write("proxy: error: Invalid log level\n")
                sys.stderr.write("See 'proxy --help'.\n")
                return 1

    configureLogging(log, logLevel)
    configureLogging(dbgp.serverBase.log, logLevel)
    try:
        proxy(dbghost, dbgport, idehost, ideport).waitForever()
    except Exception, e:
        log.exception(e)
        # some exceptions will lock the console, such as an error to bind the address
        # catching any exceptions allow us to exit without the lockup
        pass
    return 1

if __name__ == "__main__":
    main(sys.argv[1:])
